"""SessionConfig Class."""


class SessionConfig:
    """Configuration that controls the command session.

    Public Interface:
        get_config(): Get config dict, or one key/value
        set_config(): Set a config key to a value

    """

    def __init__(self):
        """Sets up a default config object."""
        self.__config = {
            "local-prompt": "$ ",
            "pexpect-delay": 0.2,  # delay required for response to be read
            "prompt-string": "$ ",
            "remote-prompt": "[ssh] $ ",
            "typing-color": "cyan",
            "typing-speed": "moderate"
        }

    def get(self, key=None):
        """Returns the config as a dict, the root dict if no key given."""

        if not key:
            return self.__config

        if key in self.__config:
            return self.__config[key]

        return None

    def set(self, key=None, value=None):
        """Sets key to value, and returns True."""

        if key in self.__config:
            self.__config[key] = value
            return True

        return False



"""TypeTastic"""

import time
import yaml
import pexpect
from pexpect import pxssh


from . import text_colors
from . import bot_handlers as bothan
from . import session_config


class Robot:
    """Robot that runs the commands."""

    # min typing speed, max typing speed, return key delay
    TypingSpeeds = {
        "slow": [0.1, 0.4, 1.0],
        "moderate": [0.05, 0.2, 0.5],
        "supersonic": [0, 0, 0]
    }

    Editors = ["vi", "vim", "emacs"]

    def __init__(self):

        self.__data = {}
        self.__config = session_config.SessionConfig()

        self.__successful_commands = 0

    def load(self, data_source):
        """Loads data either from file, dict or an array.

        Inputs:

        list of commands: ['ls', 'uptime']
        dict of commands: {"commands": ['ls', 'uptime']}
        dict of config: {"config": {'typing-color': 'red'} }
        str of file path: "myfiledata.yaml"
        """

        result = None

        if isinstance(data_source, dict):
            if "commands" in data_source or "config" in data_source:
                result = data_source

        if isinstance(data_source, list):
            result = {"commands": data_source}

        if isinstance(data_source, str):
            result = self._load_file(data_source)

        if result:
            if "commands" in result and isinstance(result["commands"], list):
                self.__data["commands"] = result["commands"]
            if "config" in result:
                # we .update() to merge into existing config defaults
                for key in result["config"]:
                    self.__config.set(key, result["config"][key])

    @staticmethod
    def setup_shell(prompt, shell="/bin/bash"):
        """Returns a pexpect spawn object.

        TODO: unit test
        """

        session = pexpect.spawn(shell, timeout=None, encoding='utf-8', echo=False)
        while True:
            session.expect_exact(['# ', '$ ', '\r\n', pexpect.EOF, pexpect.TIMEOUT])
            if not session.buffer:
                break

        session.sendline("export PS1='{0}'".format(prompt))
        time.sleep(0.2)
        while True:
            session.expect_exact([prompt, '\r\n', pexpect.EOF, pexpect.TIMEOUT])
            if not session.buffer:
                break

        return session

    def run(self):
        """Run the currently loaded commands.

        Returns:
        The number of commands with a success exit code.
        """
        typing_speed = self.__config.get("typing-speed")

        self.__successful_commands = 0  # reset this

        if "commands" in self.__data:

            # prompt = self._get_config("prompt-string")
            prompt = self.__config.get("prompt-string")
            bothan.emit_prompt(prompt)

            shell = Robot.setup_shell(prompt)

            for command in self.__data["commands"]:

                if isinstance(command, dict) and "ssh" in command:
                    ssh_conn = pxssh.pxssh()  # must be shared across all commands

                    for remote_command in command["ssh"]:

                        handler_data = {
                            "remote": ssh_conn,
                            "local": None,
                            "command": remote_command,
                            "typing_speed": self._get_typing_speeds(typing_speed),
                            "config": self.__config.get(),
                            "get_exit_status": True
                        }

                        result = self.run_task(handler_data)

                        if result:
                            self.__successful_commands += 1

                elif isinstance(command, dict) and "python3" in command:

                    for python_command in command["python3"]:

                        self.__config.set("prompt-string", ">>> ")

                        handler_data = {
                            "remote": None,
                            "local": shell,
                            "command": python_command,
                            "typing_speed": self._get_typing_speeds(typing_speed),
                            "config": self.__config.get(),
                            "get_exit_status": False
                        }

                        result = self.run_python_task(handler_data)

                        if result:
                            self.__successful_commands += 1

                else:

                    handler_data = {
                        "remote": None,
                        "local": shell,
                        "command": command,
                        "typing_speed": self._get_typing_speeds(typing_speed),
                        "config": self.__config.get(),
                        "get_exit_status": True
                    }

                    if self.run_task(handler_data):
                        self.__successful_commands += 1

            shell.close()
            time.sleep(self.__config.get("pexpect-delay") * 2)  # time to freeze frame in post
            print()  # run ends, tidy up

    @staticmethod
    def run_python_task(handler_data):
        """Run a task from handler data."""

        command = handler_data["command"]
        bothan_method = Robot._get_bothan_method(command)

        config = handler_data["config"]
        if command == "CTRL_D":
            handler_data["config"]["prompt-string"] = handler_data["config"]["local-prompt"]
            handler_data["simulated_typing"] = Robot._string_to_type(config, "^D")

        else:
            handler_data["simulated_typing"] = Robot._string_to_type(config, command)

        task_result = bothan_method(handler_data)

        prompt = handler_data["config"]["prompt-string"]
        bothan.emit_prompt(prompt)

        return task_result

    @staticmethod
    def run_task(handler_data):
        """Run a task from handler data."""

        command = handler_data["command"]
        bothan_method = Robot._get_bothan_method(command)

        config = handler_data["config"]
        handler_data["simulated_typing"] = Robot._string_to_type(config, command)

        task_result = bothan_method(handler_data)

        # trailing emit prompt, to setup the next line. pause is a special case.
        if command.startswith("PAUSE"):
            return task_result

        prompt = handler_data["config"]["prompt-string"]
        # override prompt if we are still in a remote session
        if handler_data["remote"] and not command == "exit":
            prompt = handler_data["config"]["remote-prompt"]

        bothan.emit_prompt(prompt)

        return task_result

    @staticmethod
    def _get_bothan_method(command):
        """Returns bothan method for command."""

        name = Robot._get_command_name(command)
        fn_lookup = "bot_handler_{0}".format(name)
        bothan_method = getattr(bothan, fn_lookup, bothan.bot_handler_default)

        return bothan_method

    @staticmethod
    def _get_command_name(command):
        """Returns name of the command."""
        name = command.lower()

        if " " in command:  # command has args
            (name, _) = command.split(" ", 1)

        if "/" in name:  # command has path
            (_, name) = name.rsplit("/", 1)

        return name

    def _get_data(self):
        """Return the data dict."""
        return self.__data

    def _get_config(self):
        """Return the config dict."""
        return self.__config.get()

    def _get_successful_commands(self):
        """Return the successful commands run."""
        return self.__successful_commands

    @staticmethod
    def _get_typing_speeds(speed):
        """Returns typing speeds."""
        if speed not in Robot.TypingSpeeds:
            speed = "moderate"

        return Robot.TypingSpeeds[speed]

    @staticmethod
    def _string_to_type(config, command):
        """Returns the formatted string to type."""

        color_reset = text_colors.TextColors.get_color_code("reset")
        if "typing-color" in config:
            color = text_colors.TextColors.get_color_code(config["typing-color"])
        else:
            color = ""

        command_string = "{0}{1}{2}".format(color, command, color_reset)
        return command_string

    @staticmethod
    def _load_file(inputfile):
        """Load YML file.

        Returns:
        A dictionary containing the data.
        False if data could not be loaded.
        """
        with open(inputfile, 'r') as stream:
            try:
                data = yaml.load(stream, yaml.SafeLoader)
                return data

            except yaml.YAMLError as error:
                print("load_file: {0}".format(error))
                return False



"""Handler Data Class."""


class HandlerData:
    """Handler Data Class."""

    # min typing speed, max typing speed, return key delay
    TypingSpeeds = {
        "slow": [0.1, 0.4, 1.0],
        "moderate": [0.05, 0.2, 0.5],
        "supersonic": [0, 0, 0]
    }

    def __init__(self):
        """Returns a default handler data object."""

        self.__handler_data = {
            "remote": None,
            "local": None,
            "command": None,
            "typing_speed": None,
            "config": None,
            "get_exit_status": True
        }

    def get(self, key=None):
        """Returns the handler data as a dict."""
        if not key:
            return self.__handler_data

        if key in self.__handler_data:
            return self.__handler_data[key]

        return None

    def set(self, key=None, value=None):
        """Sets key to value, and returns True."""
        if key in self.__handler_data:
            self.__handler_data[key] = value

            if key == "typing-speed":
                self.__handler_data["typing_speed"] = self._get_typing_speeds(typing_speed)
            return True

        return False

    @staticmethod
    def _get_typing_speeds(speed):
        """Returns typing speeds."""
        if speed not in HandlerData.TypingSpeeds:
            speed = "moderate"

        return HandlerData.TypingSpeeds[speed]



"""Import to reduce namespacing confusion."""

from .robot import Robot
from .handler_data import HandlerData
from .session_config import SessionConfig

# hack to ignore pyflakes warnings
assert Robot
assert HandlerData
assert SessionConfig



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="typetastic",
    version="1.1.2",
    author="Doug Bridgens",
    author_email="typetastic@far-oeuf.com",
    description="Python tool for building great screencasts, presentations, video tutorials..",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/thisdougb/typetastic",
    project_urls={
        "Bug Tracker": "https://github.com/thisdougb/typetastic/issues",
        "Source Code": "https://github.com/thisdougb/typetastic",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Topic :: Education :: Computer Aided Instruction (CAI)"
    ],
    python_requires='>=3.6',
    install_requires=[
        'getch>=1.0',
        'pexpect>=4.8.0',
        'PyYAML>=5.3.1'
    ]
)



"""Simple class for text colors."""


class TextColors:
    """Simple class for text colors."""

    colors = {
        "reset": 0,
        "black": 30,
        "red": 31,
        "green": 32,
        "yellow": 33,
        "blue": 34,
        "purple": 35,
        "cyan": 36,
        "white": 37,
    }

    @staticmethod
    def get_color_code(color_name):
        """Return the ANSI color code."""

        # defaults - cyan, not bold, not bright
        color_index = 36
        bold = 0

        for color in TextColors.colors:
            if color in color_name.lower():
                color_index = TextColors.colors[color]
                break

        if "bold" in color_name.lower():
            bold = 1

        if "bright" in color_name.lower():
            color_index += 60  # index offset

        return "\033[{0};{1}m".format(bold, color_index)



"""Handlers for commands."""

import random
import re
import sys
import time
import getch
import pexpect


def bot_handler_default(handler_data):
    """Handler for ls."""

    (speed_min, speed_max, return_key_delay) = handler_data["typing_speed"]
    simulated_typing = handler_data["simulated_typing"]
    simulate_typing(simulated_typing, speed_min, speed_max, return_key_delay)

    if handler_data["remote"]:
        return run_ssh_command(handler_data)

    return run_command(handler_data)


def bot_handler_newline(handler_data):
    # pylint: disable=unused-argument
    """Handler for newline."""
    simulate_typing("", 0, 0, 0)
    return True


def bot_handler_pause(handler_data):
    # pylint: disable=unused-argument
    """Handler for pause."""
    pause_flow()
    return True


def bot_handler_editor(handler_data):
    # pylint: disable=unused-argument
    """Handler for vi."""
    (speed_min, speed_max, return_key_delay) = handler_data["typing_speed"]
    simulated_typing = handler_data["simulated_typing"]
    simulate_typing(simulated_typing, speed_min, speed_max, return_key_delay)
    pause_flow()
    return True


# SSH
def bot_handler_ssh(handler_data):
    """SSH login."""
    if "command" in handler_data:

        (speed_min, speed_max, return_key_delay) = handler_data["typing_speed"]
        simulated_typing = handler_data["simulated_typing"]
        simulate_typing(simulated_typing, speed_min, speed_max, return_key_delay)

        command = handler_data["command"]
        ssh_conn = handler_data["remote"]

        (user, host) = parse_ssh_user_host(command)

        try:
            ssh_conn.login(host, user)
            return not ssh_conn.closed

        except pexpect.pxssh.ExceptionPxssh as error:
            print("ssh login failed: {0}".format(error))

    return False


def bot_handler_exit(handler_data):
    """SSH exit."""
    if "remote" in handler_data and handler_data["remote"]:

        (speed_min, speed_max, return_key_delay) = handler_data["typing_speed"]
        simulated_typing = handler_data["simulated_typing"]
        simulate_typing(simulated_typing, speed_min, speed_max, return_key_delay)

        ssh_conn = handler_data["remote"]
        ssh_conn.logout()
        return ssh_conn.closed

    return False


def parse_ssh_user_host(command):
    """Find user and host."""

    user_patterns = [
        r" ([a-z0-9\.]+)@",
        r"-l ([a-z0-9\.]+) "
    ]

    user = None
    for pattern in user_patterns:
        comp_pattern = re.compile(pattern)
        result = comp_pattern.search(command)
        if result:
            user = result.group(1)

    host = None
    host_patterns = [
        r"@([a-z0-9\.]+)",
        r" ([a-z0-9\.]+)$"
    ]

    for pattern in host_patterns:
        comp_pattern = re.compile(pattern)
        result = comp_pattern.search(command)
        if result:
            host = result.group(1)

    return (user, host)


# EDITORS
def bot_handler_emacs(handler_data):
    """Handler for emacs."""
    return bot_handler_editor(handler_data)


def bot_handler_vi(handler_data):
    """Handler for vi."""
    return bot_handler_editor(handler_data)


def bot_handler_crontab(handler_data):
    """Handler for vi."""
    if handler_data["command"] == "crontab -e":
        return bot_handler_editor(handler_data)
    else:
        return bot_handler_default(handler_data)


def emit_prompt(prompt):
    """Emit a prompt."""
    print(prompt, end="")
    sys.stdout.flush()


def pause_flow():
    """Creates a pause by waiting for a keypress."""
    getch.getch()


def run_python_command(handler_data):
    """Run a python interpreter command."""
    execute_command(handler_data)


def bot_handler_CTRL_D(handler_data):
    """Handler for exiting Python interactive env"""
    delay = 0.2
    session = handler_data["local"]

    (_, _, return_key_delay) = handler_data["typing_speed"]
    simulated_typing = handler_data["simulated_typing"]
    simulate_typing(simulated_typing, 0, 0, return_key_delay)

    session.sendcontrol('d')
    time.sleep(delay)
    while True:
        session.expect_exact(["$ ", '\r\n', pexpect.EOF, pexpect.TIMEOUT])
        if not session.buffer:
            break

    return True


def run_command(handler_data):
    """Run local command.
    """
    execute_command(handler_data)

    if handler_data["get_exit_status"]:
        return get_last_exit_status(handler_data)

    return True


def execute_command(handler_data):
    """Execute command."""
    delay = 0.2
    command = handler_data["command"]
    session = handler_data["local"]
    prompt = handler_data["config"]["prompt-string"]

    session.sendline(command)
    time.sleep(delay)
    while True:
        session.expect_exact([prompt, '\r\n', pexpect.EOF, pexpect.TIMEOUT])
        if session.before.rstrip():
            print(session.before.rstrip())
        if not session.buffer:
            break

    # return get_last_exit_status(handler_data)


def get_last_exit_status(handler_data):
    """This is how we get the exit status of the command."""

    delay = 0.2
    session = handler_data["local"]
    prompt = handler_data["config"]["prompt-string"]

    session.sendline("echo $?")
    time.sleep(delay)  # at least 0.2 seems to work, otherwise expect buffer is empty
    retval = False
    while True:
        session.expect_exact(["\r\n", prompt, pexpect.EOF, pexpect.TIMEOUT])
        if session.before.isdigit():
            retval = not bool(int(session.before))
        if not session.buffer:
            break

    return retval


def run_ssh_command(handler_data):
    """Run local command."""

    command = handler_data["command"]
    ssh_conn = handler_data["remote"]

    if ssh_conn:
        ssh_conn.sendline(command)
        ssh_conn.prompt()
        print(ssh_conn.before[len(command)+2:].decode("utf-8"), end="")

        ssh_conn.sendline("echo $?")
        ssh_conn.prompt()
        retval = ssh_conn.before[-3:-2].decode("utf-8")
        if retval == "0":
            return True

    # 0 = success, 1 = failure in Unix, so invert result
    return False


def simulate_typing(command, speed_min, speed_max, return_key_delay):
    """Simulates typing to stdout."""

    for char in command:
        print(char, end="")
        sys.stdout.flush()
        char_delay = random.uniform(speed_min, speed_max)
        time.sleep(char_delay)

    time.sleep(return_key_delay)
    print()  # newline required after typing command
